from arcana.core.data.space import DataSpace


class Clinical(DataSpace):
    """
    An enum used to specify the hierarchy of data trees and the "frequencies" of items
    within dataset typical of medimage research, i.e. subjects split into groups scanned
    at different timepoints (in longitudinal studies).
    """

    # Root row of the dataset
    dataset = 0b000  # singular within the dataset

    # Axes of the data space
    member = 0b001  # subjects relative to their group membership, i.e.
    # matched pairs of test and control subjects should share
    # the same member IDs.
    group = 0b010  # subject groups (e.g. test & control)
    timepoint = 0b100  # timepoints in longitudinal studies

    # Combinations
    session = 0b111  # a single session (i.e. a single timepoint of a subject)
    subject = 0b011  # uniquely identified subject within in the dataset.
    # As opposed to 'member', which specifies a subject in
    # relation to its group (i.e. one subject for each member
    # in each group). For datasets with only one study group,
    # then subject and member are equivalent
    batch = 0b110  # data from separate groups at separate timepoints
    matchedpoint = 0b101  # matched members (e.g. test & control) across
    # all groups and timepoints


class Samples(DataSpace):
    """
    The most basic data space within only one dimension
    """

    # Root row of the dataset
    dataset = 0b0  # singular within the dataset

    # Axes of the data space
    sample = 0b1



from operator import itemgetter
import click
from arcana.core.data.store import DataStore
from arcana.core.utils.serialize import ClassResolver
from arcana.core.utils.misc import get_home_dir
from arcana.core.exceptions import ArcanaUsageError
from .base import cli


@cli.group()
def store():
    pass


@store.command(
    help="""Saves the details for a new data store in the configuration
file ('~/.arcana/stores.yaml').

Arguments
---------
nickname
    The name given to the store for reference in other commands
type
    The storage class and the module it is defined in, e.g.
    `arcana.data.store.xnat:Xnat`
location
    The location of the store, e.g. server address
*varargs
    Parameters that are specific to the 'type' of storage class to be added
"""
)
@click.argument("name")
@click.argument("type")
@click.option(
    "--server",
    "-s",
    default=None,
    help="The URI of the server to connect to (if applicable)",
)
@click.option(
    "--user", "-u", default=None, help="The username to use to connect to the store"
)
@click.option(
    "--password",
    "-p",
    prompt=True,
    hide_input=True,
    help="The password to use to connect to the store",
)
@click.option(
    "--cache",
    "-c",
    default=None,
    help="The location of a cache dir to download local copies of remote data",
)
@click.option(
    "--race-condition-delay",
    "-d",
    type=int,
    help=(
        "How long to wait for changes on a incomplete download before assuming it has "
        "been interrupted, clearing it and starting again"
    ),
)
@click.option(
    "--option",
    "-o",
    nargs=2,
    multiple=True,
    metavar="<name> <value>",
    default=None,
    help="Additional key-word arguments that are passed to the store class",
)
def add(name, type, option, cache, **kwargs):
    if option is not None:
        options = dict(option)
        conflicting = set(options) & set(kwargs)
        if conflicting:
            raise ArcanaUsageError(
                f"Custom options {conflicting} conflict with in-built options, please "
                "use them instead"
            )
        kwargs.update(options)
    store_cls = ClassResolver(DataStore)(type)
    if hasattr(store_cls, "cache_dir"):
        if cache is None:
            cache = get_home_dir() / "cache" / name
        cache.mkdir(parents=True, exist_ok=True)
        kwargs["cache_dir"] = cache
    store = store_cls(name=name, **kwargs)
    store.save(name)


@store.command(
    help="""
Gives a data store saved in the config file ('~/.arcana/stores.yaml') a new
nickname.

Arguments

OLD_NICKNAME The current name of the store.
NEW_NICKNAME The new name for the store."""
)
@click.argument("old_nickname")
@click.argument("new_nickname")
def rename(old_nickname, new_nickname):
    DataStore.load(old_nickname).save(new_nickname)
    DataStore.remove(old_nickname)


@store.command(
    help="""Remove a saved data store from the config file

nickname
    The nickname the store was given when its details were saved"""
)
@click.argument("nickname")
def remove(nickname):
    DataStore.remove(nickname)


@store.command(
    help="""Refreshes credentials saved for the given store
(typically a token that expires)

nickname
    Nickname given to the store to refresh the credentials of"""
)
@click.argument("nickanme")
@click.option(
    "--user", "-u", default=None, help="The username to use to connect to the store"
)
@click.option(
    "--password",
    "-p",
    prompt=True,
    hide_input=True,
    help="The password to use to connect to the store",
)
def refresh(nickname, user, password):
    store = DataStore.load(nickname)
    if user is not None:
        store.user = user
    store.password = password
    store.save()
    DataStore.remove(nickname)
    store.save(nickname)


@store.command(help="""List available stores that have been saved""")
def ls():
    click.echo("Default stores\n---------------")
    for name, store in sorted(DataStore.singletons().items(), key=itemgetter(0)):
        click.echo(f"{name} - {ClassResolver.tostr(store, strip_prefix=False)}")
    click.echo("\nSaved stores\n-------------")
    for name, entry in DataStore.load_saved_configs().items():
        store_class = entry.pop("class")
        click.echo(f"{name} - {store_class[1:-1]}")
        for key, val in sorted(entry.items(), key=itemgetter(0)):
            click.echo(f"    {key}: {val}")



from __future__ import annotations
import typing as ty
import json
import shutil
from pathlib import Path
import attrs
import time
import yaml
from fileformats.core import FileSet, Field
from arcana.core.data.store import RemoteStore
from arcana.core.data.row import DataRow
from arcana.core.data.tree import DataTree
from arcana.core.data.entry import DataEntry
from arcana.core.data.space import DataSpace
from arcana.core.utils.misc import full_path


@attrs.define(kw_only=True)
class MockRemote(RemoteStore):
    """A simple data store to test store CLI.

    "Leaf" rows are stored in a separate sub-directories of the dataset id, with
    each part of the row id "<base-freq>=<id-part>" separated by '.', e.g.

    /path/to/dataset/leaves/a=a1.bc=b3c1.d=d7

    Under the row directory, items are stored in a directory named by their path, e.g.

    /path/to/dataset/leaves/a=a1.bc=b3c1.d=d7/scan1

    Non-leaf "rows" (e.g. higher up in the data hierarchy) are stored in a separate
    base directory and are stored by the "span" of their frequency in the dataspace, e.g.
    a row of frequency TestDataSpace.abc, would be stored at

    /path/to/dataset/nodes/a=1.b=3.c=1/
    """

    remote_dir: Path = attrs.field(converter=full_path)
    mock_delay: int = 0
    connected: bool = False
    "Mock delay used to simulate time it takes to download from remote"

    SITE_LICENSES_DIR = "LICENSE"
    METADATA_DIR = ".definition"
    LEAVES_DIR = "leaves"
    NON_LEAVES_DIR = "non-leaves"
    FIELDS_FILE = "__FIELD__"
    CHECKSUMS_FILE = "__CHECKSUMS__.json"

    #############################
    # DataStore abstractmethods #
    #############################

    def populate_tree(self, tree: DataTree):
        """
        Find all data rows for a dataset in the store and populate the
        Dataset object using its `add_leaf` method.

        Parameters
        ----------
        dataset : Dataset
            The dataset to populate with rows
        """
        with self.connection:
            self._check_connected()
            leaves_dir = self.dataset_fspath(tree.dataset_id) / self.LEAVES_DIR
            if not leaves_dir.exists():
                raise RuntimeError(
                    f"Leaves dir {leaves_dir} for flat-dir data store doesn't exist, which "
                    "means it hasn't been initialised properly"
                )
            for row_dir in self.iterdir(leaves_dir):
                ids = self.get_ids_from_row_dirname(row_dir)
                tree.add_leaf([ids[h] for h in tree.hierarchy])

    def populate_row(self, row: DataRow):
        """
        Find all data items within a data row and populate the DataRow object
        with them using the `add_fileset` and `add_field` methods.

        Parameters
        ----------
        row : DataRow
            The data row to populate with items
        """
        with self.connection:
            self._check_connected()
            try:
                row_dir = self.get_row_path(row)
            except NotInHierarchyException:
                return
            if not row_dir.exists():
                return
            for path in self.iterdir(row_dir, skip_suffixes=[".json"]):
                datatype = (
                    Field if path / self.FIELDS_FILE in path.iterdir() else FileSet
                )
                row.add_entry(
                    path=path.name,
                    datatype=datatype,
                    uri=full_path(path).relative_to(self.remote_dir),
                )

    def save_dataset_definition(
        self, dataset_id: str, definition: ty.Dict[str, ty.Any], name: str
    ):
        """Save definition of dataset within the store

        Parameters
        ----------
        dataset_id: str
            The ID/path of the dataset within the store
        definition: ty.Dict[str, Any]
            A dictionary containing the dct Dataset to be saved. The
            dictionary is in a format ready to be dumped to file as JSON or
            YAML.
        name: str
            Name for the dataset definition to distinguish it from other
            definitions for the same directory/project"""
        self._check_connected()
        definition_path = self.definition_save_path(dataset_id, name)
        definition_path.parent.mkdir(exist_ok=True)
        with open(definition_path, "w") as f:
            yaml.dump(definition, f)

    def load_dataset_definition(
        self, dataset_id: str, name: str
    ) -> ty.Dict[str, ty.Any]:
        """Load definition of a dataset saved within the store

        Parameters
        ----------
        dataset_id: str
            The ID (e.g. file-system path, XNAT project ID) of the project
        name: str
            Name for the dataset definition to distinguish it from other
            definitions for the same directory/project

        Returns
        -------
        definition: ty.Dict[str, Any]
            A dct Dataset object that was saved in the data store
        """
        self._check_connected()
        fpath = self.definition_save_path(dataset_id, name)
        if fpath.exists():
            with open(fpath) as f:
                definition = yaml.load(f, Loader=yaml.Loader)
        else:
            definition = None
        return definition

    def connect(self):
        """
        If a connection session is required to the store manage it here
        """
        self.connected = True

    def disconnect(self, session):
        """
        If a connection session is required to the store manage it here
        """
        self.connected = False

    def get_provenance(self, entry: DataEntry) -> ty.Dict[str, ty.Any]:
        self._check_connected()
        prov_path = entry.uri.with_suffix(".json")
        if prov_path.exists():
            with open(prov_path) as f:
                provenance = json.load(f)
        else:
            provenance = None
        return provenance

    def put_provenance(self, provenance: ty.Dict[str, ty.Any], entry: DataEntry):
        self._check_connected()
        prov_path = entry.uri.with_suffix(".json")
        with open(prov_path, "w") as f:
            json.dump(provenance, f)

    def create_data_tree(
        self,
        id: str,
        leaves: ty.List[ty.Tuple[str, ...]],
        hierarchy: ty.List[str],
        **kwargs,
    ):
        """reate test data within store with rows specified by row_ids

        Parameters
        ----------
        id : str
            ID of the dataset
        leaves : list[tuple[str, ...]]
            list of IDs for each leaf node to be added to the dataset. The IDs for each
            leaf should be a tuple with an ID for each level in the tree's hierarchy, e.g.
            for a hierarchy of [subject, timepoint] ->
            [("SUBJ01", "TIMEPOINT01"), ("SUBJ01", "TIMEPOINT02"), ....]
        hierarchy : list[str]
            the hierarchy of the dataset to be created
        space : type
            the dataspace of the dataset to be created
        """
        dataset_path = self.dataset_fspath(id) / self.LEAVES_DIR
        dataset_path.mkdir(parents=True)
        for ids_tuple in leaves:
            ids = dict(zip(hierarchy, ids_tuple))
            row_path = dataset_path / self.get_row_dirname_from_ids(ids, hierarchy)
            row_path.mkdir(parents=True)

    ################################
    # RemoteStore-specific methods #
    ################################

    def download_files(self, entry: DataEntry, download_dir: Path) -> Path:
        self._check_connected()
        fileset = FileSet(self.iterdir(self.entry_fspath(entry)))
        data_path = download_dir / "downloaded"
        fileset.copy(data_path, make_dirs=True, mode=fileset.CopyMode.link)
        time.sleep(self.mock_delay)
        return data_path

    def upload_files(self, cache_path: Path, entry: DataEntry):
        self._check_connected()
        entry_fspath = self.entry_fspath(entry)
        if entry_fspath.exists():
            shutil.rmtree(entry_fspath)
        entry_fspath.parent.mkdir(exist_ok=True)
        shutil.copytree(cache_path, entry_fspath)
        checksums = self.calculate_checksums(FileSet(cache_path.iterdir()))
        with open(self.remote_dir / entry.uri / self.CHECKSUMS_FILE, "w") as f:
            json.dump(checksums, f)

    def download_value(
        self, entry: DataEntry
    ) -> ty.Union[float, int, str, ty.List[float], ty.List[int], ty.List[str]]:
        """
        Retrieves a fields value

        Parameters
        ----------
        field : Field
            The field to retrieve

        Returns
        -------
        value : ty.Union[float, int, str, ty.List[float], ty.List[int], ty.List[str]]
            The value of the field
        """
        self._check_connected()
        with open(self.entry_fspath(entry) / self.FIELDS_FILE) as f:
            value = f.read()
        return value

    def upload_value(self, value, entry: DataEntry):
        self._check_connected()
        with open(self.entry_fspath(entry) / self.FIELDS_FILE, "w") as f:
            f.write(str(value))

    def create_fileset_entry(
        self, path: str, datatype: type, row: DataRow
    ) -> DataEntry:
        return self._create_entry(path=path, datatype=datatype, row=row)

    def create_field_entry(self, path: str, datatype: type, row: DataRow) -> DataEntry:
        return self._create_entry(path=path, datatype=datatype, row=row)

    def get_checksums(self, uri: str) -> ty.Dict[str, str]:
        """
        Downloads the checksum digests associated with the files in the file-set.
        These are saved with the downloaded files in the cache and used to
        check if the files have been updated on the server

        Parameters
        ----------
        uri: str
            uri of the data item to download the checksums for
        """
        fspath = self.remote_dir / uri / self.CHECKSUMS_FILE
        if not fspath.exists():
            return None
        with open(fspath) as f:
            checksums = json.load(f)
        return checksums

    def calculate_checksums(self, fileset: FileSet) -> ty.Dict[str, str]:
        """
        Downloads the checksum digests associated with the files in the file-set.
        These are saved with the downloaded files in the cache and used to
        check if the files have been updated on the server

        Parameters
        ----------
        uri: str
            uri of the data item to download the checksums for
        """
        return fileset.hash_files()

    ##################
    # Helper methods #
    ##################

    def dataset_fspath(self, dataset):
        dataset_id = (
            dataset.id if not isinstance(dataset, (str, bytes, Path)) else dataset
        )
        return self.remote_dir / dataset_id

    def entry_fspath(self, entry):
        return self.remote_dir / entry.uri

    def _create_entry(self, path: str, datatype: type, row: DataRow) -> DataEntry:
        self._check_connected()
        entry = row.add_entry(
            path=path,
            datatype=datatype,
            uri=self.get_row_path(row) / path,
        )
        self.entry_fspath(entry).mkdir(parents=True)
        return entry

    def definition_save_path(self, dataset_id, name):
        if not name:
            name = self.EMPTY_DATASET_NAME
        return self.dataset_fspath(dataset_id) / self.METADATA_DIR / (name + ".yml")

    def get_row_path(self, row: DataRow):
        dataset_fspath = self.dataset_fspath(row.dataset)
        try:
            row_path = (
                dataset_fspath
                / self.LEAVES_DIR
                / self.get_row_dirname_from_ids(row.ids, row.dataset.hierarchy)
            )
        except NotInHierarchyException:
            if not row.frequency:  # root frequency
                row_dirname = str(row.frequency)
            else:
                row_dirname = self.get_row_dirname_from_ids(
                    row.ids, [str(h) for h in row.frequency.span()]
                )
            row_path = dataset_fspath / self.NON_LEAVES_DIR / row_dirname
        return row_path

    @classmethod
    def get_row_dirname_from_ids(
        cls, ids: ty.Dict[ty.Union[str, DataSpace], str], hierarchy: ty.List[str]
    ):
        # Ensure that ID keys are strings not DataSpace enums
        ids = {str(f): i for f, i in ids.items()}
        try:
            row_dirname = ".".join(f"{h}={ids[h]}" for h in hierarchy)
        except KeyError:
            raise NotInHierarchyException
        return row_dirname

    @classmethod
    def get_ids_from_row_dirname(cls, row_dir: Path):
        parts = row_dir.name.split(".")
        return dict(p.split("=") for p in parts)

    @classmethod
    def iterdir(cls, dr, skip_suffixes=()):
        """Iterate a directory, skipping any hidden files (i.e. starting with '.'
        or any files ending in the provided suffixes)

        Parameters
        ----------
        dr : str or Path
            the directory path to iterate
        skip_suffixes : tuple, optional
            file suffixes to skip, by default ()

        Returns
        -------
        iterator
            iterator over all paths in the directory
        """
        return (
            d
            for d in Path(dr).iterdir()
            if not (
                d.name == cls.CHECKSUMS_FILE
                or d.name.startswith(".")
                or any(d.name.endswith(s) for s in skip_suffixes)
            )
        )

    def _check_connected(self):
        if not self.connected:
            raise RuntimeError("Mock data store has not been connected")


class NotInHierarchyException(Exception):
    pass



import typing as ty
from pydra.utils.hash import (
    register_serializer,
    Cache,
    bytes_repr,
    bytes_repr_sequence_contents,
)


# FIXME: Remove once this implementation has been merged into the Pydra release
@register_serializer(ty._GenericAlias)
@register_serializer(ty._SpecialForm)
@register_serializer(type)
def bytes_repr_type(klass: type, cache: Cache) -> ty.Iterator[bytes]:
    try:
        yield f"type:({klass.__module__}.{klass.__name__}".encode()
    except AttributeError:
        yield f"type:(typing.{klass._name}:(".encode()  # type: ignore
    args = ty.get_args(klass)
    if args:

        def sort_key(a):
            try:
                return a.__name__
            except AttributeError:
                return a._name

        yield b"["
        yield from bytes_repr_sequence_contents(sorted(args, key=sort_key), cache)
        yield b"]"
    yield b")"



from functools import reduce
from operator import mul
from arcana.core.cli.apply import apply_pipeline
from arcana.core.cli.derive import derive_column
from arcana.core.utils.misc import show_cli_trace
from fileformats.text import TextFile


def test_derive_cli(saved_dataset, concatenate_task, cli_runner):
    # Get CLI name for dataset (i.e. file system path prepended by 'file//')
    bp = saved_dataset.__annotations__["blueprint"]
    duplicates = 3
    # Start generating the arguments for the CLI
    # Add source to loaded dataset
    result = cli_runner(
        apply_pipeline,
        [
            saved_dataset.locator,
            "a_pipeline",
            "arcana.testing.tasks:" + concatenate_task.__name__,
            "--source",
            "file1",
            "in_file1",
            "text/text-file",
            "--source",
            "file2",
            "in_file2",
            "text/text-file",
            "--sink",
            "concatenated",
            "out_file",
            "text/text-file",
            "--parameter",
            "duplicates",
            str(duplicates),
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    # Add source column to saved dataset
    result = cli_runner(
        derive_column, [saved_dataset.locator, "concatenated", "--plugin", "serial"]
    )
    assert result.exit_code == 0, show_cli_trace(result)
    sink = saved_dataset.add_sink("concatenated", TextFile)
    assert len(sink) == reduce(mul, bp.dim_lengths)
    fnames = ["file1.txt", "file2.txt"]
    if concatenate_task.__name__.endswith("reverse"):
        fnames = [f[::-1] for f in fnames]
    expected_contents = "\n".join(fnames * duplicates)
    for item in sink:
        with open(item) as f:
            contents = f.read()
        assert contents == expected_contents



from arcana.core.data.space import DataSpace


class TestDataSpace(DataSpace):
    """Dummy data dimensions for ease of testing"""

    # Per dataset
    __ = 0b0000

    # Basis
    a = 0b1000
    b = 0b0100
    c = 0b0010
    d = 0b0001

    # Secondary combinations
    ab = 0b1100
    ac = 0b1010
    ad = 0b1001
    bc = 0b0110
    bd = 0b0101
    cd = 0b0011

    # Tertiary combinations
    abc = 0b1110
    abd = 0b1101
    acd = 0b1011
    bcd = 0b0111

    # Leaf rows
    abcd = 0b1111



from arcana.core.analysis.mark import (  # noqa: F401
    analysis,
    column,
    parameter,
    subanalysis,
    pipeline,
    switch,
    check,
    inherit,
    map_from,
    value_of,
    is_provided,
)



from pathlib import Path
import cloudpickle as cp
from pydra import mark, Workflow
from pydra.utils.hash import hash_single, Cache
from arcana.core.data.set.base import Dataset
from arcana.core.utils.serialize import asdict, fromdict


def test_dataset_asdict_roundtrip(dataset):

    dct = asdict(dataset, omit=["store", "id"])
    undct = fromdict(dct, store=dataset.store, id=dataset.id)
    assert isinstance(dct, dict)
    assert "store" not in dct
    del dataset.__annotations__["blueprint"]
    assert dataset == undct


def test_dataset_pickle(dataset: Dataset, tmp_dir: Path):
    fpath = tmp_dir / "dataset.pkl"
    with fpath.open("wb") as fp:
        cp.dump(dataset, fp)
    with fpath.open("rb") as fp:
        reloaded = cp.load(fp)
    assert dataset == reloaded


def test_dataset_in_workflow_pickle(dataset: Dataset, tmp_dir: Path):

    # Create the outer workflow to link the analysis workflow with the
    # data row iteration and store connection rows
    wf = Workflow(name="test", input_spec=["a"])

    wf.add(func(a=wf.lzin.a, b=2, dataset=dataset, name="test_func"))

    wf.set_output(("c", wf.test_func.lzout.c))

    wf.pickle_task()


@mark.task
@mark.annotate({"a": int, "b": int, "dataset": Dataset, "return": {"c": int}})
def func(a, b, dataset):
    return a + b


def test_dataset_bytes_hash(dataset):

    hsh = hash_single(dataset, Cache({}))
    # Check hashing is stable
    assert hash_single(dataset, Cache({})) == hsh



from __future__ import annotations
import typing as ty
import itertools
from abc import ABCMeta, abstractmethod
from pathlib import Path
import tempfile
import shutil
import logging
import glob
import decimal
from copy import deepcopy
import zipfile
import attrs
from fileformats.core import FileSet, Field
from fileformats.generic import Directory
from fileformats.text import TextFile
from fileformats.archive import Zip
from fileformats.field import Text as TextField, Decimal, Boolean, Integer, Array
from fileformats.serialization import Json
from fileformats.testing import (
    MyFormatGz,
    MyFormatGzX,
    MyFormatX,
    MyFormat,
    ImageWithHeader,
    YourFormat,
    Xyz,
)
from arcana.core.data.row import DataRow
from arcana.core.data.space import DataSpace
from arcana.core.utils.misc import path2varname, set_cwd
from arcana.core.exceptions import ArcanaError
from arcana.core.data.store import DataStore
from .space import TestDataSpace

logger = logging.getLogger("arcana")


@attrs.define(kw_only=True)
class EntryBlueprint(metaclass=ABCMeta):

    path: str
    datatype: type = attrs.field()
    row_frequency: ty.Optional[str] = None
    ids: ty.List[str] = None  # the list of row IDs to create the blueprint in
    alternative_datatypes: ty.List[type] = attrs.field(factory=list)

    @datatype.validator
    def datatype_validator(self, _, datatype):
        if datatype is None:
            raise ValueError("datatype cannot be None")

    @abstractmethod
    def make_item(self, **kwargs):
        pass

    def make_entry(self, row: DataRow, **kwargs):
        if self.ids and row.id not in self.ids:
            return
        item = self.make_item(**kwargs)
        logger.debug("Creating entry at %s in %s", self.path, row)
        entry = row.dataset.store.create_entry(
            path=self.path,
            datatype=self.datatype,
            row=row,
        )
        logger.debug("Putting %s at %s", item, entry)
        row.dataset.store.put(item, entry)


@attrs.define(kw_only=True)
class FileSetEntryBlueprint(EntryBlueprint):

    filenames: ty.List[str]

    def make_item(
        self,
        source_data: ty.Optional[Path] = None,
        source_fallback: bool = False,
        escape_source_name: bool = True,
    ) -> FileSet:
        """For use in test routines, this classmethod creates a simple text file,
        zip file or nested directory at the given path

        Parameters
        ----------
        source_data : Path, optional
            path to a directory containing source data to use instead of the dummy
            data
        source_fallback : bool
            whether to fall back to the generated file if fname isn't in the source
            data dir
        escape_source_name : bool
            whether to escape the source name or simple use the file name of the source

        Returns
        -------
        FileSet
            the created fileset
        """
        tmp_dir = Path(tempfile.mkdtemp())
        out_paths = []
        for fname in self.filenames:
            out_path = None
            if source_data is not None:
                src_path = source_data.joinpath(*fname.split("/"))
                fspaths = [Path(f) for f in glob.glob(str(src_path))]
                if fspaths:
                    for fspath in fspaths:
                        if escape_source_name:
                            parts = fname.split(".")
                            out_fname = (
                                path2varname(parts[0]) + "." + ".".join(parts[1:])
                            )
                        else:
                            out_fname = Path(fname).name
                        out_path = tmp_dir / out_fname
                        if fspath.is_dir():
                            shutil.copytree(fspath, out_path)
                        else:
                            shutil.copyfile(fspath, out_path, follow_symlinks=True)
                elif not source_fallback:
                    raise ArcanaError(
                        f"Couldn't find {fname} in source data directory {source_data}"
                    )
            if out_path is None:
                out_path = tmp_dir / fname
                next_part = fname
                if next_part.endswith(".zip"):
                    next_part = next_part.strip(".zip")
                next_path = Path(next_part)
                # Make double dir
                if next_part.startswith("doubledir"):
                    (tmp_dir / next_path).mkdir(exist_ok=True)
                    next_part = "dir"
                    next_path /= next_part
                if next_part.startswith("dir"):
                    (tmp_dir / next_path).mkdir(exist_ok=True)
                    next_part = "test.txt"
                    next_path /= next_part
                if not next_path.suffix:
                    next_path = next_path.with_suffix(".txt")
                if next_path.suffix == ".json":
                    contents = '{"a": 1.0}'
                else:
                    contents = fname
                with open(tmp_dir / next_path, "w") as f:
                    f.write(contents)
                if fname.endswith(".zip"):
                    with zipfile.ZipFile(out_path, mode="w") as zfile, set_cwd(tmp_dir):
                        zfile.write(next_path)
                    (tmp_dir / next_path).unlink()
            out_paths.append(out_path)
        item = self.datatype(out_paths)
        logger.debug("Created %s item", item)
        return item


@attrs.define(kw_only=True)
class FieldEntryBlueprint(EntryBlueprint):

    value: ty.Any
    expected_value: ty.Any = None

    def make_item(self, **kwargs) -> Field:
        return self.datatype(self.value)

    def __attrs_post_init__(self):
        if self.expected_value is None:
            self.expected_value = self.value


@attrs.define(slots=False, kw_only=True)
class TestDatasetBlueprint:

    space: ty.Type[DataSpace]
    hierarchy: ty.List[str]
    dim_lengths: ty.List[int]  # size of layers a-d respectively
    entries: ty.List[EntryBlueprint] = attrs.field(factory=list)
    derivatives: ty.List[EntryBlueprint] = attrs.field(factory=list)
    id_patterns: ty.Dict[str, str] = attrs.field(factory=dict)
    include: ty.Dict[str, ty.Union[str, ty.List[str]]] = attrs.field(factory=dict)
    exclude: ty.Dict[str, ty.Union[str, ty.List[str]]] = attrs.field(factory=dict)

    DEFAULT_NUM_ACCESS_ATTEMPTS = 300
    DEFAULT_ACCESS_ATTEMPT_INTERVAL = 1.0  # secs

    def make_dataset(
        self,
        store: DataStore,
        dataset_id: str,
        name: ty.Optional[str] = None,
        source_data: ty.Optional[Path] = None,
        metadata: ty.Optional[ty.Dict[str, ty.Any]] = None,
        **kwargs,
    ):
        """For use in tests, this method creates a test dataset from the provided
        blueprint

        Parameters
        ----------
        store: DataStore
            the store to make the dataset within
        dataset_id : str
            the ID of the project/directory within the store to create the dataset
        name : str, optional
            the name to give the dataset. If provided the dataset is also saved in the
            datastore
        source_data : Path, optional
            path to a directory containing source data to use instead of the dummy
            data
        **kwargs
            passed through to create_dataset
        """
        if metadata is None:
            metadata = {}
        orig_type = metadata.get("type", "derivative")
        metadata["type"] = "in-construction"
        with store.connection:
            logger.debug(
                "Creating test dataset in %s at %s from %s", store, dataset_id, self
            )
            if self.id_patterns:
                kwargs = {"id_patterns": self.id_patterns}
            dataset = store.create_dataset(
                id=dataset_id,
                leaves=self.all_ids,
                name=name,
                hierarchy=self.hierarchy,
                space=self.space,
                metadata=metadata,
                include=self.include,
                exclude=self.exclude,
                **kwargs,
            )
        with store.connection:
            logger.debug(
                "Adding entries to test dataset for: %s",
                dataset.rows(frequency=max(self.space)),
            )
            for row in dataset.rows(frequency=max(self.space)):
                self.make_entries(row, source_data=source_data)
            dataset.metadata.type = orig_type
            dataset.save()
        dataset.__annotations__["blueprint"] = self
        logger.debug("Successfully created test dataset at %s in %s", dataset_id, store)
        return dataset

    def translate_to(self, data_store: DataStore) -> "TestDatasetBlueprint":
        """Translates the blueprint so that it matches the default space and hierarchy
        of the data store (if applicable)

        Parameters
        ----------
        data_store : DataStore
            the data store to get the defaults for

        Returns
        -------
        blueprint : TestDatasetBlueprint
        """
        # Create copy of the blueprint
        blueprint = deepcopy(self)
        try:
            blueprint.space = data_store.DEFAULT_SPACE
        except AttributeError:
            space = TestDataSpace
        else:
            try:
                blueprint.hierarchy = data_store.DEFAULT_HIERARCHY
            except AttributeError:
                if space.ndim > self.space.ndim:
                    raise RuntimeError(
                        f"cannot translate hierarchy as from {self.space} to {space} "
                        "as it has more dimensions"
                    )
                # Translate frequencies into new space
                blueprint.hierarchy = [
                    space.union(*f.span()[-space.ndim :]) for f in self.hierarchy
                ]
                # Drop frequencies that mapped onto same value
                blueprint.hierarchy = [
                    h
                    for i, h in enumerate(blueprint.hierarchy)
                    if i == 0 or h != blueprint.hierarchy[i - 1]
                ]
                blueprint.dim_lengths = self.dim_lengths[-len(blueprint.hierarchy) :]
        return blueprint

    # def access_dataset(
    #     self,
    #     store: DataStore,
    #     dataset_id: str,
    #     name: ty.Optional[str] = None,
    #     max_num_attempts: int = DEFAULT_NUM_ACCESS_ATTEMPTS,
    #     attempt_interval: float = DEFAULT_ACCESS_ATTEMPT_INTERVAL,
    # ):
    #     """For data stores with significant latency, this method can be used to reuse
    #     test datasets between tests

    #     Parameters
    #     ----------
    #     store : DataStore
    #         the data store to access the dataset from
    #     dataset_id : str
    #         the ID of the dataset to access
    #     name : str, optional
    #         the name of the dataset
    #     max_num_attempts: int, optional
    #         the maximum number of attempts to try to access
    #     attempt_interval: float, optional
    #         the time (in secs) between each attempt

    #     Returns
    #     -------
    #     Dataset
    #         the accessed dataset
    #     """
    #     num_attempts = 0
    #     while num_attempts < max_num_attempts:
    #         try:
    #             dataset = store.load_dataset(dataset_id, name=name)
    #         except KeyError:
    #             pass
    #         else:
    #             if dataset.metadata.type != "in-construction":
    #                 break
    #         logger.info(
    #             "Waiting for test dataset '%s' to finish being constructed",
    #             dataset_id,
    #         )
    #         time.sleep(attempt_interval)
    #         num_attempts += 1
    #     if num_attempts >= max_num_attempts:
    #         wait_time = max_num_attempts * attempt_interval
    #         raise RuntimeError(
    #             f"Could not access {dataset_id} dataset in {store} after waiting "
    #             f"{wait_time}, something may have gone wrong in the construction process"
    #         )
    #     dataset.__annotations__["blueprint"] = self
    #     return dataset

    def make_entries(
        self,
        row: DataRow,
        **kwargs,
    ):
        """Creates the actual data in the store, from the provided blueprint, which
        can be used to run test routines against

        Parameters
        ----------
        store
            the store to create the test dataset in
        dataset_id : str
            the ID of the dataset to create
        name : str
            the name of the dataset
        **kwargs
            passed directly through to the EntryBlueprint.create_item method
        """
        logger.debug("making entries for %s: %s", row, self.entries)
        for entry_bp in self.entries:
            entry_bp.make_entry(row, **kwargs)

    @property
    def all_ids(self):
        """Iterate all leaves of the data tree specified by the test blueprint and yield
        ID tuples corresponding to the IDs of each leaf node"""
        for id_tple in itertools.product(*(list(range(d)) for d in self.dim_lengths)):
            base_ids = dict(zip(self.space.axes(), id_tple))
            ids = {}
            for layer in self.hierarchy:
                ids[layer] = "".join(
                    f"{b}{base_ids[b]}" for b in self.space[layer].span()
                )
            yield tuple(ids[h] for h in self.hierarchy)


TEST_DATASET_BLUEPRINTS = {
    "full": TestDatasetBlueprint(  # dataset name
        space=TestDataSpace,
        hierarchy=["a", "b", "c", "d"],
        dim_lengths=[2, 3, 4, 5],
        entries=[
            FileSetEntryBlueprint(
                path="file1", datatype=TextFile, filenames=["file1.txt"]
            ),
            FileSetEntryBlueprint(
                path="file2", datatype=MyFormatGz, filenames=["file2.my.gz"]
            ),
            FileSetEntryBlueprint(path="dir1", datatype=Directory, filenames=["dir1"]),
            FieldEntryBlueprint(
                path="textfield",
                row_frequency="abcd",
                datatype=TextField,
                value="sample-text",
            ),  # Derivatives to insert
            FieldEntryBlueprint(
                path="booleanfield",
                row_frequency="c",
                datatype=Boolean,
                value="no",
                expected_value=False,
            ),  # Derivatives to insert
        ],
        derivatives=[
            FileSetEntryBlueprint(
                path="deriv1",
                row_frequency="abcd",
                datatype=TextFile,
                filenames=["file1.txt"],
            ),  # Derivatives to insert
            FileSetEntryBlueprint(
                path="deriv2",
                row_frequency="c",
                datatype=Directory,
                filenames=["dir"],
            ),
            FileSetEntryBlueprint(
                path="deriv3",
                row_frequency="bd",
                datatype=TextFile,
                filenames=["file1.txt"],
            ),
            FieldEntryBlueprint(
                path="integerfield",
                row_frequency="c",
                datatype=Integer,
                value=99,
            ),
            FieldEntryBlueprint(
                path="decimalfield",
                row_frequency="bd",
                datatype=Decimal,
                value="33.3333",
                expected_value=decimal.Decimal("33.3333"),
            ),
            FieldEntryBlueprint(
                path="arrayfield",
                row_frequency="bd",
                datatype=Array[Integer],
                value=[1, 2, 3, 4, 5],
            ),
        ],
    ),
    "one_layer": TestDatasetBlueprint(
        space=TestDataSpace,
        hierarchy=["abcd"],
        dim_lengths=[1, 1, 1, 5],
        entries=[
            FileSetEntryBlueprint(
                path="file1",
                datatype=MyFormatGzX,
                filenames=["file1.my.gz", "file1.json"],
                alternative_datatypes=[MyFormatGz, Json],
            ),
            FileSetEntryBlueprint(
                path="file2",
                datatype=MyFormatX,
                filenames=["file2.my", "file2.json"],
                alternative_datatypes=[MyFormat, Json],
            ),
        ],
        derivatives=[
            FileSetEntryBlueprint(
                path="deriv1",
                row_frequency="abcd",
                datatype=Json,
                filenames=["file1.json"],
            ),
            FileSetEntryBlueprint(
                path="deriv2",
                row_frequency="bc",
                datatype=Xyz,
                filenames=["file1.x", "file1.y", "file1.z"],
            ),
            FileSetEntryBlueprint(
                path="deriv3",
                row_frequency="__",
                datatype=YourFormat,
                filenames=["file1.yr"],
            ),
        ],
    ),
    "skip_single": TestDatasetBlueprint(
        space=TestDataSpace,
        hierarchy=["a", "bc", "d"],
        dim_lengths=[2, 1, 2, 3],
        entries=[
            FileSetEntryBlueprint(
                path="doubledir1", datatype=Directory, filenames=["doubledir1"]
            ),
            FileSetEntryBlueprint(
                path="doubledir2", datatype=Directory, filenames=["doubledir2"]
            ),
        ],
        derivatives=[
            FileSetEntryBlueprint(
                path="deriv1",
                row_frequency="ad",
                datatype=Json,
                filenames=["file1.json"],
            )
        ],
    ),
    "skip_with_inference": TestDatasetBlueprint(
        space=TestDataSpace,
        hierarchy=["bc", "ad"],
        dim_lengths=[2, 3, 2, 4],
        id_patterns={
            "a": r"ad::a(\d+)d\d+",
            "b": r"bc::b(\d+)c\d+",
            "c": r"bc::b\d+c(\d+)",
            "d": r"ad::a\d+d(\d+)",
        },
        entries=[
            FileSetEntryBlueprint(
                path="file1",
                datatype=ImageWithHeader,
                filenames=["file1.hdr", "file1.img"],
            ),
            FileSetEntryBlueprint(
                path="file2", datatype=YourFormat, filenames=["file2.yr"]
            ),
        ],
    ),
    "redundant": TestDatasetBlueprint(
        space=TestDataSpace,
        hierarchy=[
            "abc",
            "abcd",
        ],  # e.g. XNAT where session ID is unique in project but final layer is organised by timepoint
        dim_lengths=[3, 4, 5, 6],
        id_patterns={
            "a": r"abc::a(\d+)b\d+c\d+",
            "b": r"abc::a\d+b(\d+)c\d+",
            "c": r"abc::a\d+b\d+c(\d+)",
            "d": r"abcd::a\d+b\d+c\d+d(\d+)",
        },
        entries=[
            FileSetEntryBlueprint(
                path="doubledir", datatype=Directory, filenames=["doubledir"]
            ),
            FileSetEntryBlueprint(
                path="file1", datatype=Xyz, filenames=["file1.x", "file1.y", "file1.z"]
            ),
        ],
        derivatives=[
            FileSetEntryBlueprint(
                path="deriv1",
                row_frequency="d",
                datatype=Json,
                filenames=["file1.json"],
            )
        ],
    ),
    "concatenate_test": TestDatasetBlueprint(
        space=TestDataSpace,
        hierarchy=[
            "abcd"
        ],  # e.g. XNAT where session ID is unique in project but final layer is organised by timepoint
        dim_lengths=[1, 1, 1, 2],
        entries=[
            FileSetEntryBlueprint(
                path="file1", datatype=TextFile, filenames=["file1.txt"]
            ),
            FileSetEntryBlueprint(
                path="file2", datatype=TextFile, filenames=["file2.txt"]
            ),
        ],
    ),
    "concatenate_zip_test": TestDatasetBlueprint(
        space=TestDataSpace,
        hierarchy=[
            "abcd"
        ],  # e.g. XNAT where session ID is unique in project but final layer is organised by timepoint
        dim_lengths=[1, 1, 1, 1],
        entries=[
            FileSetEntryBlueprint(path="file1", datatype=Zip, filenames=["file1.zip"]),
            FileSetEntryBlueprint(path="file2", datatype=Zip, filenames=["file2.zip"]),
        ],
    ),
}


GOOD_DATASETS = ["full", "one_layer", "skip_single", "skip_with_inference", "redundant"]


EXTENSION_DATASET_BLUEPRINTS = {
    "complete": TestDatasetBlueprint(  # dataset name
        space=TestDataSpace,
        hierarchy=["a", "b", "c", "d"],
        dim_lengths=[2, 2, 2, 2],
        entries=[
            FileSetEntryBlueprint(
                path="file1", datatype=TextFile, filenames=["file.txt"]
            ),
            FileSetEntryBlueprint(
                path="file2", datatype=MyFormatGz, filenames=["file.my.gz"]
            ),
            FileSetEntryBlueprint(
                path="file3",
                datatype=MyFormatGzX,
                filenames=["file.my.gz", "file.json"],
            ),
            FileSetEntryBlueprint(path="dir1", datatype=Directory, filenames=["dir1"]),
            FieldEntryBlueprint(
                path="textfield",
                row_frequency="abcd",
                datatype=TextField,
                value="sample-text",
            ),  # Derivatives to insert
            FieldEntryBlueprint(
                path="booleanfield",
                row_frequency="c",
                datatype=Boolean,
                value="no",
                expected_value=False,
            ),  # Derivatives to insert
        ],
        derivatives=[
            FileSetEntryBlueprint(
                path="deriv1",
                row_frequency="abcd",
                datatype=TextFile,
                filenames=["file1.txt"],
            ),  # Derivatives to insert
            FileSetEntryBlueprint(
                path="deriv2",
                row_frequency="c",
                datatype=Directory,
                filenames=["dir"],
            ),
            FileSetEntryBlueprint(
                path="deriv3",
                row_frequency="bd",
                datatype=TextFile,
                filenames=["file1.txt"],
            ),
            FieldEntryBlueprint(
                path="integerfield",
                row_frequency="c",
                datatype=Integer,
                value=99,
            ),
            FieldEntryBlueprint(
                path="decimalfield",
                row_frequency="bd",
                datatype=Decimal,
                value="33.3333",
                expected_value=decimal.Decimal("33.3333"),
            ),
            FieldEntryBlueprint(
                path="arrayfield",
                row_frequency="bd",
                datatype=Array[Integer],
                value=[1, 2, 3, 4, 5],
            ),
        ],
    ),
}

SIMPLE_DATASET = TestDatasetBlueprint(  # dataset name
    space=TestDataSpace,
    hierarchy=["abcd"],
    dim_lengths=[2, 2, 2, 2],
    entries=[
        FileSetEntryBlueprint(path="file1", datatype=TextFile, filenames=["file.txt"]),
        FieldEntryBlueprint(path="field1", datatype=TextField, value="a field"),
    ],
)



import typing as ty
import re
from enum import Enum
from arcana.core.utils.serialize import ClassResolver
from arcana.core.utils.misc import classproperty


class DataSpace(Enum):
    """
    Base class for all "data space" enums. DataSpace enums specify
    the relationships between rows of a dataset.

    For example in imaging studies, scannings sessions are typically organised
    by analysis group (e.g. test & control), membership within the group (i.e
    matched subjects) and time-points (for longitudinal studies). We can
    visualise the rows arranged in a 3-D grid along the `group`, `member`, and
    `timepoint` dimensions. Note that datasets that only contain one group or
    time-point can still be represented in the same space, and just be of
    depth=1 along those dimensions.

    All dimensions should be included as members of a DataSpace subclass
    enum with orthogonal binary vector values, e.g.

        member = 0b001
        group = 0b010
        timepoint = 0b100

    In this space, an imaging session row is uniquely defined by its member,
    group and timepoint ID. The most commonly present dimension should be given
    the least frequent bit (e.g. imaging datasets will not always have
    different groups or time-points but will always have different members
    (equivalent to subjects when there is one group).

    In addition to the data items stored in the data rows for each session,
    some items only vary along a particular dimension of the grid. The
    "row_frequency" of these rows can be specified using the "basis" members
    (i.e. member, group, timepoint) in contrast to the `session` row_frequency,
    which is the combination of all three

        session = 0b111

    Additionally, some data is stored in aggregated rows that across a plane
    of the grid. These frequencies should also be added to the enum (all
    combinations of the basis frequencies must be included) and given intuitive
    names if possible, e.g.

        subject = 0b011 - uniquely identified subject within in the dataset.
        batch = 0b110 - separate group+timepoint combinations
        matchedpoint = 0b101 - matched members and time-points aggregated across groups

    Finally, for items that are singular across the whole dataset there should
    also be a dataset-wide member with value=0:

        dataset = 0b000
    """

    def __str__(self):
        return self.name

    @classmethod
    def leaf(cls):
        return max(cls)

    @classmethod
    def axes(cls):
        return cls.leaf().span()

    @classproperty
    def ndim(self):
        return len(self.axes())

    def span(self):
        """Returns the basis dimensions in the data tree that the given
        enum-member projects into.

        For example in `Clinical` data trees, the following frequencies can
        be decomposed into the following basis dims:

            dataset -> []
            group -> [group]
            member -> [member]
            timepoint -> [timepoint]
            subject -> [group, member]
            batch -> [timepoint, group]
            matchedpoint -> [timepoint, member]
            session -> [timepoint, group, member]
        """
        # Check which bits are '1', and append them to the list of levels
        cls = type(self)
        return [cls(b) for b in sorted(self.nonzero_bits(), reverse=True)]

    def nonzero_bits(self):
        v = self.value
        nonzero = []
        while v:
            w = v & (v - 1)
            nonzero.append(w ^ v)
            v = w
        return nonzero

    def __iter__(self):
        "Iterate over bit string"
        bit = (max(type(self)).value + 1) >> 1
        while bit > 0:
            yield bool(self.value & bit)
            bit >>= 1

    def is_basis(self):
        return len(self._nonzero_bits()) == 1

    def __eq__(self, other):
        return self.value == other.value

    def __lt__(self, other):
        return self.value < other.value

    def __le__(self, other):
        return self.value <= other.value

    def __xor__(self, other):
        return type(self)(self.value ^ other.value)

    def __and__(self, other):
        return type(self)(self.value & other.value)

    def __or__(self, other):
        return type(self)(self.value | other.value)

    def __invert__(self):
        return type(self)(~self.value)

    def __hash__(self):
        return self.value

    def __bool__(self):
        return bool(self.value)

    def bin(self):
        return bin(self.value)

    @classmethod
    def union(cls, freqs: ty.Sequence[Enum]):
        "Returns the union between data frequency values"
        union = cls(0)
        for f in freqs:
            union |= f if isinstance(f, cls) else cls[f]
        return union

    @classmethod
    def default(cls):
        return max(cls)

    def is_parent(self, child, if_match=False):
        """Checks to see whether the current frequency is a "parent" of the
        other data frequency, i.e. all the base row_frequency of self appear in
        the "child".

        Parameters
        ----------
        child : DataSpace
            The data frequency to check parent/child relationship with
        if_match : bool
            Treat matching frequencies as "parents" of each other

        Returns
        -------
        bool
            True if self is parent of child
        """
        return ((self & child) == self) and (child != self or if_match)

    def tostr(self):
        return f"{ClassResolver.tostr(self, strip_prefix=False)}[{str(self)}]"

    @classmethod
    def fromstr(cls, s):
        match = re.match(r"(.*)\[([^\]]+)\]", s)
        if match is None:
            raise ValueError(
                f"'{s}' is not a string of the format <data-space-enum>[<value>]"
            )
        class_loc, val = match.groups()
        space = ClassResolver(cls)(class_loc)
        return space[val] if not isinstance(space, str) else s

    @classproperty
    def SUBPACKAGE(cls):
        """Cannot be a regular class attribute because then DataSpace won't be able to
        be extended"""
        return "data"



import pytest
from pathlib import Path
import docker
import docker.errors
from arcana.core.utils.misc import show_cli_trace
from arcana.core.cli.deploy import make_app, install_license
from arcana.testing.deploy.licenses import (
    get_pipeline_image,
    make_dataset,
    ORG,
    REGISTRY,
    IMAGE_VERSION,
    LICENSE_CONTENTS,
    LICENSE_NAME,
    LICENSE_INPUT_FIELD,
    LICENSE_OUTPUT_FIELD,
    LICENSE_PATH_PARAM,
    LICENSE_INPUT_PATH,
    LICENSE_OUTPUT_PATH,
)


def test_buildtime_license(license_file, run_prefix: str, work_dir: Path, cli_runner):

    # Create pipeline
    image_name = f"license-buildtime-{run_prefix}"
    image_tag = f"{REGISTRY}/{ORG}/{image_name}:{IMAGE_VERSION}"

    root_dir = work_dir / ORG
    root_dir.mkdir()
    spec_file = root_dir / (image_name + ".yaml")

    LICENSE_PATH = "/path/to/licence.txt"

    pipeline_image = get_pipeline_image(LICENSE_PATH)
    pipeline_image.name = image_name
    pipeline_image.licenses[0].store_in_image = True
    pipeline_image.save(spec_file)

    build_dir = work_dir / "build"
    dataset_dir = work_dir / "dataset"
    make_dataset(dataset_dir)

    result = cli_runner(
        make_app,
        args=[
            str(root_dir),
            "common:App",
            "--build-dir",
            str(build_dir),
            "--license",
            LICENSE_NAME,
            str(license_file),
            "--use-local-packages",
            "--install-extras",
            "test",
            "--raise-errors",
            "--registry",
            REGISTRY,
        ],
    )

    assert result.exit_code == 0, show_cli_trace(result)

    assert result.stdout.strip().splitlines()[-1] == image_tag

    args = (
        "dirtree///dataset "
        f"--input {LICENSE_INPUT_FIELD} '{LICENSE_INPUT_PATH}' "
        f"--output {LICENSE_OUTPUT_FIELD} '{LICENSE_OUTPUT_PATH}' "
        f"--parameter {LICENSE_PATH_PARAM} '{LICENSE_PATH}' "
        f"--plugin serial "
        f"--raise-errors "
    )

    dc = docker.from_env()
    try:
        result = dc.containers.run(
            image_tag,
            args,
            volumes=[
                f"{str(dataset_dir)}:/dataset:rw",
            ],
            remove=False,
            stdout=True,
            stderr=True,
        )
    except docker.errors.ContainerError as e:
        logs = e.container.logs().decode("utf-8")
        raise RuntimeError(
            f"Running {image_tag} failed with args = {args}" f"\n\nlogs:\n{logs}",
        )


def test_site_runtime_license(license_file, work_dir, arcana_home, cli_runner):

    # build_dir = work_dir / "build"
    dataset_dir = work_dir / "dataset"

    LICENSE_PATH = work_dir / "license_location"

    pipeline_image = get_pipeline_image(LICENSE_PATH)

    # Save it into the new home directory
    dataset = make_dataset(dataset_dir)

    result = cli_runner(install_license, args=[LICENSE_NAME, str(license_file)])
    assert result.exit_code == 0, show_cli_trace(result)

    pipeline_image.command.execute(
        dataset.locator,
        input_values={LICENSE_INPUT_FIELD: LICENSE_INPUT_PATH},
        output_values={LICENSE_OUTPUT_FIELD: LICENSE_OUTPUT_PATH},
        parameter_values={LICENSE_PATH_PARAM: LICENSE_PATH},
        work_dir=work_dir / "pipeline",
        raise_errors=True,
        plugin="serial",
        loglevel="info",
    )


def test_dataset_runtime_license(license_file, run_prefix, work_dir, cli_runner):

    # build_dir = work_dir / "build"
    dataset_dir = work_dir / "dataset"

    dataset = make_dataset(dataset_dir)

    LICENSE_PATH = work_dir / "license_location"
    pipeline_image = get_pipeline_image(LICENSE_PATH)

    result = cli_runner(
        install_license,
        args=[
            LICENSE_NAME,
            str(license_file),
            dataset.locator,
        ],
    )

    assert result.exit_code == 0, show_cli_trace(result)

    pipeline_image.command.execute(
        dataset.locator,
        input_values={LICENSE_INPUT_FIELD: LICENSE_INPUT_PATH},
        output_values={LICENSE_OUTPUT_FIELD: LICENSE_OUTPUT_PATH},
        parameter_values={LICENSE_PATH_PARAM: LICENSE_PATH},
        work_dir=work_dir / "pipeline",
        raise_errors=True,
        plugin="serial",
        loglevel="info",
    )


@pytest.fixture
def license_file(work_dir) -> Path:
    license_src = work_dir / "license_file.txt"

    with open(license_src, "w") as f:
        f.write(LICENSE_CONTENTS)

    return license_src
